@@ -1,6 +1,6 @@
 /* 
    socket handling interface
-   Copyright (C) 1999-2010, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1999-2021, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -42,7 +42,8 @@ NE_BEGIN_DECLS
 #define NE_SOCK_RESET (-4)
 /* Secure connection was closed without proper SSL shutdown. */
 #define NE_SOCK_TRUNC (-5)
-/* Reserved value -   (-6) */
+/* Retry operation later. */
+#define NE_SOCK_RETRY (-6)
 
 /* ne_socket represents a TCP socket. */
 typedef struct ne_socket_s ne_socket;
@@ -112,7 +113,9 @@ typedef enum {
 ne_inet_addr *ne_iaddr_make(ne_iaddr_type type, const unsigned char *raw);
 
 /* Compare two network address objects i1 and i2; returns zero if they
- * are equivalent or non-zero otherwise.  */
+ * are equivalent or non-zero otherwise. For an IPv6 literal, the
+ * scope is ignored by this function if set, so must be compared
+ * seperately if required. */
 int ne_iaddr_cmp(const ne_inet_addr *i1, const ne_inet_addr *i2);
 
 /* Return the type of the given network address object. */
@@ -138,6 +141,12 @@ int ne_iaddr_reverse(const ne_inet_addr *ia, char *buf, size_t bufsiz);
  * non-NULL, return value must be freed using ne_iaddr_free. */
 ne_inet_addr *ne_iaddr_parse(const char *addr, ne_iaddr_type type);
 
+/* Set the scope_id for an IPv6 address from 'scope'. */
+int ne_iaddr_set_scope(ne_inet_addr *ia, const char *scope);
+
+/* Returns the scope_id for an IPv6 address. */
+char *ne_iaddr_get_scope(const ne_inet_addr *ia);
+
 /* Destroy a network address object created using ne_iaddr_make or
  * ne_iaddr_parse. */
 void ne_iaddr_free(ne_inet_addr *addr);
@@ -208,7 +217,7 @@ int ne_sock_fullwritev(ne_socket *sock, const struct ne_iovec *vector,
  * At most 'len' bytes are read (including the NUL terminator).
  * Returns:
  * NE_SOCK_* on error,
- * >0 number of bytes read (including NUL terminator)
+ * >0 number of bytes read (excluding NUL terminator)
  */
 ssize_t ne_sock_readline(ne_socket *sock, char *buffer, size_t len);
 
@@ -227,10 +236,31 @@ int ne_sock_fd(const ne_socket *sock);
  * must be destroyed by caller using ne_iaddr_free. */
 ne_inet_addr *ne_sock_peer(ne_socket *sock, unsigned int *port);
 
-/* Close the socket and destroy the socket object.  If SSL is in use
- * for the socket, a closure alert is sent to initiate a clean
- * shutdown, but this function does not wait for the peer's response.
- * Returns zero on success, or non-zero on failure. */
+/* Flags for ne_sock_shutdown():  */
+#define NE_SOCK_RECV (1)
+#define NE_SOCK_SEND (2)
+#define NE_SOCK_BOTH (3)
+
+/* Shut down the socket in one or both directions, without destroying
+ * the socket object.  Flags must be one of NE_SOCK_RECV/SEND/BOTH.
+ * For a non-TLS socket, performs the directional shutdown according
+ * to flags.
+ * For a TLS socket:
+ * - if flags are NE_SOCK_SEND or NE_SOCK_BOTH, sends the TLS
+ *   close_notify.  Returns NE_SOCK_RETRY if the TLS connection has
+ *   not been closed by the peer.
+ * - if flags are NE_SOCK_RECV, returns NE_SOCK_RETRY if the 
+ *   TLS close_notify has not been closed by the peer.
+ * In NE_SOCK_SEND or NE_SOCK_BOTH is specified, and the bidirectional
+ * TLS shutdown has completed, the TCP shutdown will also be completed 
+ * as for a non-TLS socket. 
+*/
+int ne_sock_shutdown(ne_socket *sock, unsigned int flags);
+
+/* Close the socket if it is open, and destroy the socket object.  If
+ * SSL is in use for the socket, a closure alert is sent to initiate a
+ * clean shutdown, but this function does not wait for the peer's
+ * response.  Returns zero on success, or non-zero on failure. */
 int ne_sock_close(ne_socket *sock);
 
 /* Return current error string for socket. */
@@ -276,6 +306,10 @@ int ne_sock_sessid(ne_socket *sock, unsigned char *buf, size_t *buflen);
  * freed by the caller. */
 char *ne_sock_cipher(ne_socket *sock);
 
+/* Return the SSL/TLS protocol version used for socket 'sock', or
+ * NE_SSL_PROTO_UNSPEC if SSL/TLS is not in use for the socket. */
+enum ne_ssl_protocol ne_sock_getproto(ne_socket *sock);
+
 /* SOCKS proxy protocol version: */
 enum ne_sock_sversion {
     NE_SOCK_SOCKSV4 = 0,
