@@ -1,6 +1,6 @@
 /* 
    URI manipulation routines.
-   Copyright (C) 1999-2021, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1999-2025, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -42,7 +42,7 @@
 #include "ne_alloc.h"
 #include "ne_uri.h"
 
-/* URI ABNF from RFC 3986: */
+/* Character classes defined from URI ABNF from RFC 3986. */
 
 #define PS (0x0001) /* "+" */
 #define PC (0x0002) /* "%" */
@@ -68,6 +68,10 @@
 #define OT (0x4000) /* others */
 /* UNUSED  (0x8000) .. only remaining bit. */
 
+/* ALL defined characters classes, note that 0x8000 is an unused class
+ * bit so is not (currently) included. */
+#define URI_ALL (0x7FFF)
+
 #define URI_ALPHA (AL)
 #define URI_DIGIT (DG)
 #define URI_NONURI (OT)
@@ -91,8 +95,17 @@
 /* fragment == query */
 #define URI_FRAGMENT URI_QUERY
 
-/* any characters which should be path-escaped: */
-#define URI_ESCAPE ((URI_GENDELIM & ~(FS)) | URI_SUBDELIM | OT | PC)
+/* Used previously for the ne_path_escape() rule, or NE_PATH_NONRES
+ * flag for ne_path_escapef(): */
+#define URI_OLD_ESCAPE ((URI_GENDELIM & ~(FS)) | URI_SUBDELIM | OT | PC)
+
+/* Characters which should be pct-encoded are the INVERSE of the
+ * SEGCHAR class. Used now by ne_path_escape() and the NE_PATH_NONPC
+ * rule. */
+#define URI_ESCAPE ((URI_ALL & ~(URI_SEGCHAR)) | PC)
+
+/* Maximum allowed port number. */
+#define MAX_PORT (65535)
 
 static const unsigned short uri_chars[256] = {
 /* 0xXX    x0      x2      x4      x6      x8      xA      xC      xE     */
@@ -220,23 +233,30 @@ int ne_uri_parse(const char *uri, ne_uri *parsed)
 
         parsed->host = ne_strndup(s, p - s);
 
-        if (p != pa && p + 1 != pa) {
-            p++;
+        /* Iff p and pa (=> path-abempty) differ, the optional port
+         * section is present and parsed here: */
+        if (p != pa) {
+            unsigned int port = 0;
 
-            s = p;
-            /* => s = port */
+            if (*p++ != ':') return -1;
 
-            while (p < pa) {
-                if (!(uri_lookup(*p) & URI_DIGIT))
-                    return -1;
+            /* => p = port */
 
-                p++;
-            }
+            /* port = *DIGIT
+             *
+             * Note: port can be the empty string, in which case now:
+             * p == pa and port is parsed as 0, as desired. */
+            while (p < pa && port <= MAX_PORT && (uri_lookup(*p) & URI_DIGIT) != 0)
+                port = 10*port + *p++-'0';
 
-            parsed->port = atoi(s);
+            /* If p did not reach pa there was some non-digit present
+             * or the integer was too large, so fail. */
+            if (p != pa || port > MAX_PORT) return -1;
+
+            parsed->port = port;
         }
         
-        s = pa;
+        s = pa; /* Next, parse path-abempty */
     }
 
     /* => s = path-abempty / path-absolute / path-rootless
@@ -493,7 +513,7 @@ char *ne_path_unescape(const char *uri)
 
 char *ne_path_escape(const char *path)
 {
-    return ne_path_escapef(path, NE_PATH_NONRES);
+    return ne_path_escapef(path, NE_PATH_NONPC);
 }
 
 char *ne_path_escapef(const char *path, unsigned int flags)
@@ -503,8 +523,9 @@ char *ne_path_escapef(const char *path, unsigned int flags)
     size_t count = 0;
     unsigned short mask = 0;
 
-    if (flags & NE_PATH_NONRES) mask |= URI_ESCAPE;
+    if (flags & NE_PATH_NONRES) mask |= URI_OLD_ESCAPE;
     if (flags & NE_PATH_NONURI) mask |= URI_NONURI;
+    if (flags & NE_PATH_NONPC) mask |= URI_ESCAPE;
 
     for (pnt = (const unsigned char *)path; *pnt != '\0'; pnt++) {
         count += path_escape_ch(*pnt, mask);
@@ -603,9 +624,7 @@ char *ne_uri_unparse(const ne_uri *uri)
         if (uri->port > 0
             && (!uri->scheme 
                 || ne_uri_defaultport(uri->scheme) != uri->port)) {
-            char str[20];
-            ne_snprintf(str, 20, ":%d", uri->port);
-            ne_buffer_zappend(buf, str);
+            ne_buffer_snprintf(buf, 20, ":%d", uri->port);
         }
     }
 
