@@ -29,6 +29,9 @@
 #ifdef HAVE_ERRNO_H
 #include <errno.h> /* for the ENOENT definitions in str_errors */
 #endif
+#ifdef HAVE_LIMITS_H
+#include <limits.h>
+#endif
 
 #include "ne_string.h"
 #include "ne_utils.h"
@@ -36,7 +39,7 @@
 #include "tests.h"
 
 #undef ONCMP
-#define ONCMP(a,b) ONV(strcmp(a, b), \
+#define ONCMP(a,b) ONV(!a || strcmp(a, b), \
 		       ("result was [%s] not [%s]", a, b))
 
 static int simple(void) {
@@ -497,7 +500,7 @@ static int printing(void)
 
         ret = ne_snprintf(buf, ts[n].pass, "%s", ts[n].in);
         
-        ONCMP(buf, ts[n].out);
+        ONCMP(ts[n].out, buf);
         ONV(ret != ts[n].ret, 
             ("got return value %" NE_FMT_SIZE_T " not %" NE_FMT_SIZE_T,
              ret, ts[n].ret));
@@ -622,6 +625,7 @@ static int qappend(void)
         { "b", 2, "b\\x00" },
         { "alpha\0alpha", 11, "alpha\\x00alpha" },
         { "a\tb", 3, "a\\x09b" },
+        { "foo\x7f" "bar", 7, "foo\\x7fbar" },
         { NULL }
     };
     unsigned n;
@@ -650,6 +654,190 @@ static int qappend(void)
     return OK;
 }
 
+static char *test_vstrhash(unsigned int flags, ...)
+{
+    va_list ap;
+    char *rv;
+
+    va_start(ap, flags);
+    rv = ne_vstrhash(flags, ap);
+    va_end(ap);
+
+    return rv;
+}
+
+#define TEST1 "abcdbcdecdefdefgefghfghighijhijkijkljklmklmnlmnomnopnopq"
+#define TEST1_SHA "248d6a61d20638b8e5c026930c3e6039a33ce45964ff2167f6ecedd419db06c1"
+
+#define ONVEC(args, expect) \
+    do { char *h = ne_strhash args; ONCMP(h, expect); ne_free(h); } while (0)
+
+static int strhash(void)
+{
+    ONN("zero flags must return NULL", ne_strhash(0, "", NULL) != NULL);
+    ONN("zero flags must return NULL for vstrhash", test_vstrhash(0, "", NULL) != NULL);
+
+    ONN("no alg flags must return NULL", ne_strhash(NE_HASH_COLON, "", NULL) != NULL);
+    ONN("no alg flags must return NULL", ne_strhash(NE_HASH_SPACE, "", NULL) != NULL);
+    
+    ONVEC((NE_HASH_MD5, "", NULL), "d41d8cd98f00b204e9800998ecf8427e");
+    ONVEC((NE_HASH_MD5, "foo", "ba", "r", NULL), "3858f62230ac3c915f300c664312c63f");
+    ONVEC((NE_HASH_MD5|NE_HASH_SPACE, "foo", "ba", "r", NULL), "38 58 f6 22 30 ac 3c 91 5f 30 0c 66 43 12 c6 3f");
+
+    return OK;
+}
+
+static int strhash_sha_256(void)
+{
+    char *p = ne_strhash(NE_HASH_SHA256, "", NULL);
+    if (p == NULL) {
+        t_context("SHA-2-256 not supported");
+        return SKIP;
+    }
+    ne_free(p);
+
+    ONVEC((NE_HASH_SHA256, TEST1, NULL), TEST1_SHA);
+    ONVEC((NE_HASH_SHA256, "foobar", "foo", "bar", "f", "oobar", NULL),
+          "d173c93898d3ca8455a4526e0af2a1aee9b91c8ec19adac16e6e8be2da09436c");
+
+    return OK;
+}
+
+/* NIST examples from https://csrc.nist.gov/CSRC/media/Projects/Cryptographic-Standards-and-Guidelines/documents/examples/SHA512.pdf */
+#define TEST1_512 "abc"
+#define TEST1_512_MDC "dd:af:35:a1:93:61:7a:ba:cc:41:73:49:ae:20:41:31:12:e6:fa:4e:89:a9:7e:a2:0a:9e:ee:e6:4b:55:d3:9a:21:92:99:2a:27:4f:c1:a8:36:ba:3c:23:a3:fe:eb:bd:45:4d:44:23:64:3c:e8:0e:2a:9a:c9:4f:a5:4c:a4:9f"
+#define TEST2_512_1 "abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijklmnoijklmnopjklmnopqklmnopqrl"
+#define TEST2_512_2 "mnopqrsmnopqrstnopqrstu"
+#define TEST2_512_MD "8e959b75dae313da8cf4f72814fc143f8f7779c6eb9f7fa17299aeadb6889018501d289e4900f7e4331b99dec4b5433ac7d329eeb6dd26545e96e55b874be909"
+
+
+/* NIST examples from https://csrc.nist.gov/CSRC/media/Projects/Cryptographic-Standards-and-Guidelines/documents/examples/SHA512_256.pdf */
+#define TEST1_512_256 "abc"
+#define TEST1_512_256_MD "53048e2681941ef99b2e29b76b4c7dabe4c2d0c634fc6d46e0e2f13107e7af23"
+#define TEST2_512_256_1 "abcdefghbcdefghicdefghijdefghijkefghijklfghijklmghijklmnhijkl"
+#define TEST2_512_256_2 "mnoijklmnopjklmnopqklmnopqrlmnopqrsmnopqrstnopqrstu"
+#define TEST2_512_256_MD "3928e184fb8690f840da3988121d31be65cb9d3ef83ee6146feac861e19b563a"
+#define TEST2_512_256_MDC "39:28:e1:84:fb:86:90:f8:40:da:39:88:12:1d:31:be:65:cb:9d:3e:f8:3e:e6:14:6f:ea:c8:61:e1:9b:56:3a"
+
+static int strhash_sha_512(void)
+{
+    char *p = ne_strhash(NE_HASH_SHA512, "", NULL);
+
+    if (p == NULL) {
+        t_context("SHA-2-512 not supported");
+        return SKIP;
+    }
+    ne_free(p);
+
+    ONVEC((NE_HASH_SHA512|NE_HASH_COLON, TEST1_512, NULL), TEST1_512_MDC);
+    ONVEC((NE_HASH_SHA512, TEST2_512_1, TEST2_512_2, NULL), TEST2_512_MD);
+
+    return OK;
+}
+
+static int strhash_sha_512_256(void)
+{
+    char *p = ne_strhash(NE_HASH_SHA512_256, "", NULL);
+
+    if (p == NULL) {
+        t_context("SHA-2-512/256 not supported");
+        return SKIP;
+    }
+    ne_free(p);
+
+    ONVEC((NE_HASH_SHA512_256, TEST1_512_256, NULL), TEST1_512_256_MD);
+    ONVEC((NE_HASH_SHA512_256, TEST2_512_256_1, TEST2_512_256_2, NULL), TEST2_512_256_MD);
+    ONVEC((NE_HASH_SHA512_256|NE_HASH_COLON, TEST2_512_256_1, TEST2_512_256_2, NULL),
+          TEST2_512_256_MDC);
+
+    return OK;
+}
+
+static int strparam(void)
+{
+    static const struct {
+        const char *charset, *lang;
+        const char *value;
+        const char *expect;
+    } ts[] = {
+        { "UTF-8", NULL, "foobar", NULL },
+        { "UTF-8", NULL, "foo12345bar", NULL },
+        { "UTF-8", NULL, "foo@bar", "UTF-8''foo%40bar" },
+        { "UTF-8", NULL, "foo bar", "UTF-8''foo%20bar" },
+        { "iso-8859-1", "en", "\xA3 rates", "iso-8859-1'en'%a3%20rates" },
+        { "UTF-8", NULL, "£ and € rates", "UTF-8''%c2%a3%20and%20%e2%82%ac%20rates" },
+        { NULL }
+    };
+    unsigned n;
+
+    for (n = 0; ts[n].charset; n++) {
+        char *act = ne_strparam(ts[n].charset, ts[n].lang, (const unsigned char *)ts[n].value);
+
+        if (ts[n].expect == NULL) {
+            ONV(act != NULL, ("expected NULL output for '%s', got '%s'",
+                              ts[n].value, act));
+        }
+        else {
+            ONCMP(act, ts[n].expect);
+            ne_free(act);
+        }
+    }
+
+    return OK;
+}
+
+static int strhextoul(void)
+{
+    static const struct {
+        const char *input;
+        unsigned long expect;
+        int error; /* errno */
+        const char *endp;
+    } *t, ts[] = {
+        { "0x0", ULONG_MAX, EINVAL },
+        { "0X1", ULONG_MAX, EINVAL },
+        { "+1", ULONG_MAX, EINVAL },
+        { "-0", ULONG_MAX, EINVAL },
+        { "+0x1", ULONG_MAX, EINVAL },
+        { "+0X1", ULONG_MAX, EINVAL },
+        { "", ULONG_MAX, EINVAL },
+        { "1", 1, 0, "" },
+        { " 10", ULONG_MAX, EINVAL, " 10" },
+        { "0000010 ", 16, 0, " " },
+        { "4242", 16962, 0 },
+        { "4242zZZz", 16962, 0, "zZZz" },
+        { "cAfEBeEf", 3405692655 },
+#if SIZEOF_LONG == 8
+        { "10000000000000000", ULONG_MAX, ERANGE },
+        { "100000000", 0x100000000, 0 },
+#elif SIZEOF_LONG == 4
+        { "100000000", ULONG_MAX, ERANGE },
+#endif
+        { NULL }
+    };
+    unsigned n;
+
+    for (n = 0; ts[n].input; n++) {
+        unsigned long actual;
+        const char *endp = "(unset)", **endpp;
+        int errnum;
+
+        t = ts + n;
+
+        endpp = t->endp ? &endp : NULL;
+        errno = ENOENT;
+        actual = ne_strhextoul(t->input, endpp);
+        errnum = errno;
+        ONV(errnum != t->error,
+            ("got errno %d not %d for [%s]", errnum, t->error, t->input));
+        ONV(actual != t->expect,
+            ("got %lu not %lu for [%s]", actual, t->expect, t->input));
+        if (endpp) ONCMP(*endpp, t->endp);
+    }
+
+    return OK;
+}
+
 ne_test tests[] = {
     T(simple),
     T(buf_concat),
@@ -678,6 +866,12 @@ ne_test tests[] = {
     T(casencmp),
     T(buf_print),
     T(qappend),
+    T(strhash),
+    T(strhash_sha_256),
+    T(strhash_sha_512),
+    T(strhash_sha_512_256),
+    T(strparam),
+    T(strhextoul),
     T(NULL)
 };
 
